library(igraph)

`getParents` <- function(poset, i)

{
	
	allParents <- which(poset[, i] == 1)
	
	sort(allParents)
	
}

`matrixPower` <- function (X, n){
	## unfortunately R doesn't have a matrix power function like MATLAB.
	## This function is taken from packages Biodem and is originally called mtx.exp.
	if (n != round(n)) {
		n <- round(n)
		warning("rounding exponent `n' to", n)
	}
	phi <- diag(nrow = nrow(X))
	pot <- X
	while (n > 0) {
		if (n%%2) 
			phi <- phi %*% pot
		n <- n%/%2
		pot <- pot %*% pot
	}
	return(phi)
}



`transitiveClosure` <-
		function(poset){
	## Returns the transitive closure of a set of relations,
	## i.e., the minimal poset containing the relations
	p <- nrow(poset)
	for (i in 2:p){
		poset <- poset + matrixPower(poset, i)
	}
	poset <- apply(poset, 2, function(x) as.numeric(x>0))
	return(poset)
}

`generateData` <- function(poset, thetas, eps, N)
{
	n <- length(thetas)
	data <- matrix(0, N, n)
	
	parents_list <- list()
	
	
	for(j in 1:n)
	{
		parents_list[[j]] <- getParents(poset, j)
	}
	
	g <- graph.adjacency( poset )
	topo_sorted <- topological.sort(g)
	
#print(paste("after for:", Sys.time()))
	
	for(i in 1:N)
	{
		for(j in topo_sorted)
		{
			if(length(parents_list[[j]]) == 0 || prod(data[i, parents_list[[j]] ] ) == 1)
			{
				data[i, j] = rbinom(1, 1, thetas[j])
			} 
		}
		
		eps_noise <- rbinom(n, 1, eps)
		
		data[i, ] = (data[i, ] + eps_noise) %%2
	}
	
	data
}

