/*
 * bayes_dt_cbn.h
 *
 *  Created on: Jan 9, 2012
 *      Author: Thomas Sakoparnig
 */

#ifndef BAYES_DT_CBN_H_
#define BAYES_DT_CBN_H_

#include "dt_cbn.h"

/**
 * computes the likelihood of the data
 * @param eps
 * @param M
 * @param theta
 * @param D
 * @param N_u
 * @param Prob
 * @param cond_err_prob
 * @return likelihood
 */
long double compute_likelihood( double eps, model* M, double* theta, data* D, int N_u, double* Prob, double** cond_err_prob ) {

	int i,j;
	int m = M->m;

	long double likelihood, likelihood_d;

	likelihood = 1;
	//double loglik = 0;
	//product over observed genotypes (i.e. data)
	for( i=0;i<N_u;i++ ) {
		//sum over valid genotypes
		likelihood_d = 0;
		for( j=0;j<m;j++ ) {
			likelihood_d += Prob[j] * cond_err_prob[j][i];
		}
		likelihood *= power( likelihood_d, D[i].count );
		//loglik += D[i].count * log( likelihood_d );
	}
	//printf("\n likelihood: %20Lg\n", likelihood);
	//printf("\n loglik: %20Lg\n", logl( likelihood ) );
	return likelihood;
}


/**
 * unnorm P[M,theta,eps|D]
 * @param eps
 * @param M
 * @param theta
 * @param D
 * @param N_u
 * @return unnormalized posterior
 */
long double compute_unnomr_posterior( double eps, model*M, double* theta, data* D, int N_u ) {

	int m = M->m;
	int n = M->n;

	double* Prob = get_double_array(m);
	double* theta_exit = get_double_array(m);
	compute_theta_exit( theta, M, theta_exit);
	compute_all_cbn_prob( theta, M, Prob, theta_exit );
	free(theta_exit);
	double** cond_err_prob = get_double_matrix(m, N_u);
	compute_cond_error_prob( M, D, N_u, eps, cond_err_prob );
	long double lik;
	lik = compute_likelihood(eps, M, theta, D, N_u, Prob, cond_err_prob );
	free(Prob);
	int i;
	free_double_matrix( cond_err_prob, m );

	double t=1;
	double alpha, beta;
	for( i=0;i<n;i++ ) {
		//TODO: treat alpha and beta like parameters not like f***-given constants
		alpha = 0.00001;
		beta = 0.00001;
		alpha=beta=1;
		t *= gsl_ran_beta_pdf( theta[i], alpha, beta);
	}
	//TODO: treat alpha and beta like parameters not like f***-given constants
	double eps_pdf = gsl_ran_beta_pdf( eps, 5, 30 );
	long double posterior;
	posterior = lik*t*eps_pdf;

	return posterior;
}

void relocate_whole_theta( double* theta, double* theta_p, int n ) {
	int i;
	double alpha,beta,var,x;
	for( i=0;i<n;i++ ) {
		x = theta[i];
		//TODO: set alpha and beta!!!
		beta = 1;
		var = x * (1 - x) / ( beta / ( 1 - x) + 1 );
		alpha = x * ( x * ( 1 - x ) / var - 1 );
		alpha = beta = 1;
		theta_p[i] = gsl_ran_beta (RNG, alpha, beta);
	}
}

void relocate_theta_i( double* theta, double* theta_p, int n, int i ) {
	double alpha,beta,var,x;
	x = theta[i];
	//TODO: set alpha and beta!!!
	//beta = 1;
	//var = x * (1 - x) / ( beta / ( 1 - x) + 1 );
	//alpha = x * ( x * ( 1 - x ) / var - 1 );
	alpha = beta = 1;
	theta_p[i] = gsl_ran_beta (RNG, alpha, beta);
}

double compute_theta_transition_prob( double* theta, double* theta_p, int n ) {
	double t = 1;
	int i;
	double alpha,beta,var,x;
	for( i=0;i<n;i++ ) {
		x = theta[i];
		//TODO: set alpha and beta!!!
		//beta = 1;
		//var = x * (1 - x) / ( beta / ( 1 - x) + 1 );
		//alpha = x * ( x * ( 1 - x ) / var - 1 );
		alpha = beta = 1;
		t *= gsl_ran_beta_pdf(theta_p[i], alpha, beta);
	}
	return t;
}

//double relocate_epsilon( double epsilon ) {
//	//TODO: set and eval alpha and beta properly!!!
//	return gsl_ran_beta( RNG, 5, 30 );
//}
//
//double compute_epsilon_transition_prob( double epsilon ) {
//	//TODO: set and eval alpha and beta properly!!!
//	return gsl_ran_beta_pdf( epsilon, 5, 30 );
//}


void valid_new_cover_relation_matrix( model *M, int** V ) {
	int n = M->n;
	int i,j,k,l;
	int** T = get_int_matrix(n, n);
	int** C = get_int_matrix(n, n);

	for( i=0;i<n;i++ ) {
		for( k=0;k<n;k++ ) {
			C[i][k] = 0;
			T[i][k] = 0;
		}
	}
	int stat;
	transitive_closure( M->P, T, n );

	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( !T[i][j] && i!=j ) {
				T[i][j] = 1;
				stat = reduce_to_cover_relations( T, n, C );
				//check changes
				if( stat == 1 && !C[i][j] ) {
					for( k=0;k<n;k++ ) {
						for( l=0;l<n;l++ ) {
							if( C[k][l] == 1 && M->P[k][l] == 1 ) {
								T[i][j] = 0;
							}
						}
					}
					V[i][j] = T[i][j];
				}
				//cycle!!
				else if( stat == 2) {
					T[i][j] = 0;
				}
				else if( stat == 0 ){
					V[i][j] = 1;
				}
			}

			for( l=0;l<n;l++ ) {
				for( k=0;k<n;k++ ) {
					T[l][k] = 0;
					C[l][k] = 0;
				}
			}
			transitive_closure( M->P, T, n );
		}
	}
	free_int_matrix( T, n );
	free_int_matrix( C, n );
}

int propose_new_cover_relation( model *M, double* tp, data* D, int N_u ) {
	int n = M->n;
	int i,j,k,N_compatible,N_all_comp;
	int** V = get_int_matrix(n, n);
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ )
			V[i][j] = 0;
	}
	valid_new_cover_relation_matrix( M, V );
	int valid_c = 0;
	int c = 0;
	N_all_comp = 0;

	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( V[i][j] ) {
				valid_c++;
				//M->P[i][j] = 1;
				//compatibility( D, N_u, M );
				//N_compatible = 0;
				//for (k=0; k<N_u; k++) {
				//N_compatible += D[k].is_compatible * D[k].count;
				//N_compatible = 1;
				//}
				N_compatible = 1;
				V[i][j] = N_compatible;
				N_all_comp += N_compatible;
				//M->P[i][j] = 0;
			}
		}
	}
	//TODO: what if N_compatible is 0 for a specific instert???
	if (valid_c) {
		//(*tp) = (double)1/valid_c;
		int new_index = rand()%N_all_comp;
		for( i=0;i<n;i++ ) {
			for( j=0;j<n;j++ ) {
				if( V[i][j] ) {
					c += V[i][j];
					if( new_index < c ) {
						M->P[i][j] = 1;
						(*tp) = (double)V[i][j]/N_all_comp;
						new_index = N_all_comp + 1;
					}
				}
			}
		}
	}
	else {
		free_int_matrix( V, n );
		return 0;
	}
	free_int_matrix( V, n );
	return 1;
}

double get_tp_for_remove_cover_move( model* M ) {
	int i,j;
	int delete_c = 0;
	int n = M->n;
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if ( M->P[i][j] ) {
				delete_c++;
			}
		}
	}
	return (double)1/delete_c;
}

int propose_delete_cover_relation( model *M, double* tp ) {
	int valid_c = 0;
	int c = 0;
	int i,j;
	int n = M->n;
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( M->P[i][j] ) {
				valid_c++;
			}
		}
	}
	if (valid_c) {
		(*tp) = (double)1/valid_c;
		int new_index = rand()%valid_c;
		for( i=0;i<n;i++ ) {
			for( j=0;j<n;j++ ) {
				if( M->P[i][j] ) {
					if( new_index == c ) {
						M->P[i][j] = 0;
					}
					c++;
				}
			}
		}
	}
	else {
		return 0;
	}
	return 1;
}

double get_tp_for_new_cover_move( model* M, data* D, int N_u ) {
	int i,j,N_all_comp,N_compatible,k;
	N_all_comp = 0;
	int c = 0;
	int n = M->n;
	double tp = 0;
	int** V = get_int_matrix(n, n);
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			V[i][j] = 0;
		}
	}
	valid_new_cover_relation_matrix( M, V );
	int valid_c = 0;
	N_all_comp = 0;

	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( V[i][j] ) {
				valid_c++;
				//M->P[i][j] = 1;
				//compatibility( D, N_u, M );
				//N_compatible = 0;
				//for (k=0; k<N_u; k++) {
				//N_compatible += D[k].is_compatible * D[k].count;
				//N_compatible = 1;
				//}
				//N_compatible = 1;
				//V[i][j] = N_compatible;
				//N_all_comp += N_compatible;
				//M->P[i][j] = 0;
			}
		}
	}

	if (valid_c) {
		//(*tp) = (double)1/valid_c;
		int new_index = rand()%valid_c;
		for( i=0;i<n;i++ ) {
			for( j=0;j<n;j++ ) {
				if( V[i][j] ) {
					c += V[i][j];
					if( new_index < c ) {
						tp = (double)V[i][j]/valid_c;
						new_index = valid_c + 1;
					}
				}
			}
		}
	}
	free_int_matrix( V, n );
	return tp;
}

void propose_event_exchange_move( model* M, double* tp, int *ti, int *tj ) {
	int n = M->n;
	int i,j,k;
	int* h_in = get_int_array( n );
	int* h_out = get_int_array( n );
	int h=0;
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			h += M->P[i][j];
		}
	}

	i = j = rand()%n;
	while( i==j ) {
		j = rand()%n;
	}

	int ij,ji;
	ij = M->P[i][j];
	ji = M->P[j][i];

	for( k=0;k<n;k++ ) {
		h_out[k] = M->P[i][k];
		h_in[k] = M->P[k][i];
	}
	for( k=0;k<n;k++ ) {
		if ( i != k ) {
			M->P[i][k] = M->P[j][k];
			M->P[k][i] = M->P[k][j];
		}
	}
	for( k=0;k<n;k++ ) {
		if ( j != k ) {
			M->P[j][k] = h_out[k];
			M->P[k][j] = h_in[k];
		}
	}
	M->P[i][j] = ji;
	M->P[j][i] = ij;

	(*ti) = i;
	(*tj) = j;


	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			h -= M->P[i][j];
		}
	}

	if ( h != 0 ) {
		printf("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\n");
	}

	(*tp) = 2/(double)(n*(n-1));

}


//death move followed by birth move
int propose_reincarnation_move( model* M, data* D, int N_u ) {
	double tp;
	if (! propose_delete_cover_relation( M, &tp) ) {
		return 0;
	}
	else {
		propose_new_cover_relation( M, &tp, D, N_u );
	}
	return 1;
}

void relocate_epsilon( double eps, double* epsilon_p ) {
	double alpha,beta,var,x;

	//BetaDist(2,20) is a bit broader than (5,30) used for the prior
	(*epsilon_p) = gsl_ran_beta (RNG, 2, 20);
}

double get_tp_epsilon_relocation( double epsilon_p ) {

	return gsl_ran_beta_pdf( epsilon_p, 2, 20);
}

void valid_new_transitive_closer_relation_matrix( model *M, int** V ) {
	int n = M->n;
	int i,j,k,l;
	int** T = get_int_matrix(n, n);
	int** T2 = get_int_matrix(n, n);
	int** C = get_int_matrix(n, n);
	int new_n_edges;
	int orig_n_edges = 0;
	int changes;

	for( i=0;i<n;i++ ) {
		for( k=0;k<n;k++ ) {
			C[i][k] = 0;
			T[i][k] = 0;
			T2[i][k] = 0;
		}
	}
	int stat;
	transitive_closure( M->P, T, n );
	for( l=0;l<n;l++ ) {
		for( k=0;k<n;k++ ) {
			orig_n_edges += T[l][k];
		}
	}

	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( !T[i][j] && i!=j ) {
				T[i][j] = 1;
				transitive_closure( T, T2, n );
				new_n_edges = 0;
				for( l=0;l<n;l++ ) {
					for( k=0;k<n;k++ ) {
						new_n_edges += T2[l][k];
					}

				}
				if ((new_n_edges - orig_n_edges) == 1) {
					stat = reduce_to_cover_relations( T2, n, C );
					//check changes
					if( stat == 1 && !C[i][j] ) {
						changes = 0;
						for( l=0;l<n;l++ ) {
							for( k=0;k<n;k++ ) {
								changes += abs(T2[l][k]-M->P[l][k]);
							}
						}
						if ( changes > 2 ) {
							V[i][j] = 1;
						}
					}
				}
				T[i][j] = 0;
				for( l=0;l<n;l++ ) {
					for( k=0;k<n;k++ ) {
						T2[l][k] = 0;
						C[l][k] = 0;
					}
				}
			}
		}
	}
	free_int_matrix( T, n );
	free_int_matrix( T2, n );
	free_int_matrix( C, n );
}

int propose_new_transitive_closure_relation( model *M, double* tp, data* D, int N_u ) {
	int n = M->n;
	int i,j,k,N_compatible,N_all_comp;
	int** V = get_int_matrix(n, n);
	int** T = get_int_matrix(n, n);
	int** C = get_int_matrix(n, n);
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			V[i][j] = 0;
			T[i][j] = 0;
		}
	}
	valid_new_transitive_closer_relation_matrix( M, V );
	int valid_c = 0;
	int c = 0;
	N_all_comp = 0;
	transitive_closure( M->P, T, n );

	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( V[i][j] ) {
				valid_c++;
				//M->P[i][j] = 1;
				//compatibility( D, N_u, M );
				//N_compatible = 0;
				//for (k=0; k<N_u; k++) {
				//N_compatible += D[k].is_compatible * D[k].count;
				//N_compatible = 1;
				//}
				N_compatible = 1;
				V[i][j] = N_compatible;
				N_all_comp += N_compatible;
				//M->P[i][j] = 0;
			}
		}
	}
	//TODO: what if N_compatible is 0 for a specific instert???
	if (valid_c) {
		//(*tp) = (double)1/valid_c;
		int new_index = rand()%N_all_comp;
		for( i=0;i<n;i++ ) {
			for( j=0;j<n;j++ ) {
				if( V[i][j] ) {
					c += V[i][j];
					if( new_index < c ) {
						T[i][j] = 1;
						reduce_to_cover_relations( T, n, C );
						(*tp) = (double)V[i][j]/N_all_comp;
						new_index = N_all_comp + 1;
					}
				}
			}
		}
		for( i=0;i<n;i++ ) {
			for( j=0;j<n;j++ ) {
				M->P[i][j] = T[i][j];
			}
		}
	}
	else {
		free_int_matrix( V, n );
		free_int_matrix( C, n );
		free_int_matrix( T, n );
		return 0;
	}
	free_int_matrix( V, n );
	free_int_matrix( C, n );
	free_int_matrix( T, n );
	return 1;
}
void valid_delete_transitive_closure_relation_matrix( model *M, int** V ) {
	int n = M->n;
	int i,j,k,l;
	int** T = get_int_matrix(n, n);
	int** T2 = get_int_matrix(n, n);
	int** C = get_int_matrix(n, n);
	int changes = 0;

	for( i=0;i<n;i++ ) {
		for( k=0;k<n;k++ ) {
			C[i][k] = 0;
			T[i][k] = 0;
			T2[i][k] = 0;
		}
	}
	int stat;
	transitive_closure( M->P, T, n );
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( T[i][j] && i!=j ) {
				T[i][j] = 0;
				transitive_closure( T, T2, n );
				if ( !T2[i][j] ) {
					stat = reduce_to_cover_relations( T2, n, C );
					//check changes
					if( stat == 1 && !C[i][j] ) {
						changes = 0;
						for( l=0;l<n;l++ ) {
							for( k=0;k<n;k++ ) {
								changes += abs(T2[l][k]-M->P[l][k]);
							}
						}
						if ( changes > 2 ) {
							V[i][j] = 1;
						}
					}
				}
				T[i][j] = 1;
				for( l=0;l<n;l++ ) {
					for( k=0;k<n;k++ ) {
						T2[l][k] = 0;
						C[l][k] = 0;
					}
				}
			}
		}
	}
	free_int_matrix( T, n );
	free_int_matrix( T2, n );
	free_int_matrix( C, n );
}

int propose_delete_transitive_closure_relation( model *M, double* tp, data* D, int N_u ) {

	int n = M->n;
	int i,j,k,N_compatible,N_all_comp;
	int** V = get_int_matrix(n, n);
	int** T = get_int_matrix(n, n);
	int** C = get_int_matrix(n, n);
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ )
			V[i][j] = 0;
	}
	valid_delete_transitive_closure_relation_matrix( M, V );
	int valid_c = 0;
	int c = 0;
	N_all_comp = 0;

	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( V[i][j] ) {
				valid_c++;
				//M->P[i][j] = 1;
				//compatibility( D, N_u, M );
				//N_compatible = 0;
				//for (k=0; k<N_u; k++) {
				//N_compatible += D[k].is_compatible * D[k].count;
				//N_compatible = 1;
				//}
				//N_compatible = 1;
				//V[i][j] = N_compatible;
				//N_all_comp += N_compatible;
				//M->P[i][j] = 0;
			}
		}
	}
	//TODO: what if N_compatible is 0 for a specific instert???
	if (valid_c) {
		//(*tp) = (double)1/valid_c;
		int new_index = rand()%valid_c;
		for( i=0;i<n;i++ ) {
			for( j=0;j<n;j++ ) {
				if( V[i][j] ) {
					c += V[i][j];
					if( new_index < c ) {
						T[i][j] = 1;
						reduce_to_cover_relations( T, n, C );
						(*tp) = (double)V[i][j]/valid_c;
						new_index = valid_c + 1;
					}
				}
			}
		}
		for( i=0;i<n;i++ ) {
			for( j=0;j<n;j++ ) {
				M->P[i][j] = T[i][j];
			}
		}
	}
	else {
		free_int_matrix( V, n );
		free_int_matrix( C, n );
		free_int_matrix( T, n );
		return 0;
	}
	free_int_matrix( V, n );
	free_int_matrix( C, n );
	free_int_matrix( T, n );
	return 1;
}

double get_tp_for_delete_transitive_closure_relation_move( model* M ) {
	int n = M->n;
	int i,j,k,N_compatible,N_all_comp;
	int** V = get_int_matrix(n, n);
	int** T = get_int_matrix(n, n);
	int** C = get_int_matrix(n, n);
	double tp = 0;
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ )
			V[i][j] = 0;
	}
	valid_delete_transitive_closure_relation_matrix( M, V );
	int valid_c = 0;
	int c = 0;
	N_all_comp = 0;

	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( V[i][j] ) {
				valid_c++;
				//M->P[i][j] = 1;
				//compatibility( D, N_u, M );
				//N_compatible = 0;
				//for (k=0; k<N_u; k++) {
				//N_compatible += D[k].is_compatible * D[k].count;
				//N_compatible = 1;
				//}
				//N_compatible = 1;
				//V[i][j] = N_compatible;
				//N_all_comp += N_compatible;
				//M->P[i][j] = 0;
			}
		}
	}
	//TODO: what if N_compatible is 0 for a specific instert???
	if (valid_c) {
		tp = (double)1/valid_c;

	}
	else {
		free_int_matrix( V, n );
		free_int_matrix( C, n );
		free_int_matrix( T, n );
		return 0;
	}
	free_int_matrix( V, n );
	free_int_matrix( C, n );
	free_int_matrix( T, n );
	return tp;
}

double get_tp_for_new_transitive_closure_relation_move( model* M ) {
	int n = M->n;
	int i,j,k,N_compatible,N_all_comp;
	int** V = get_int_matrix(n, n);
	int** T = get_int_matrix(n, n);
	int** C = get_int_matrix(n, n);
	double tp = 0;
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ )
			V[i][j] = 0;
	}
	valid_new_transitive_closer_relation_matrix( M, V );
	int valid_c = 0;
	int c = 0;
	N_all_comp = 0;

	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( V[i][j] ) {
				valid_c++;
				//M->P[i][j] = 1;
				//compatibility( D, N_u, M );
				//N_compatible = 0;
				//for (k=0; k<N_u; k++) {
				//N_compatible += D[k].is_compatible * D[k].count;
				//N_compatible = 1;
				//}
				//N_compatible = 1;
				//V[i][j] = N_compatible;
				//N_all_comp += N_compatible;
				//M->P[i][j] = 0;
			}
		}
	}
	//TODO: what if N_compatible is 0 for a specific instert???
	if (valid_c) {
		tp = (double)1/valid_c;

	}
	else {
		free_int_matrix( V, n );
		free_int_matrix( C, n );
		free_int_matrix( T, n );
		return 0;
	}
	free_int_matrix( V, n );
	free_int_matrix( C, n );
	free_int_matrix( T, n );
	return tp;
}

/**
 *
 * @param eps IN epsilon
 * @param M
 * @param theta
 * @param D
 * @param N_u
 * @param burn_in
 * @param number_samples
 */
void start_MH( double eps, model* M, double* theta, data* D, int N_u, int burn_in, int number_samples, int record_ith ) {

	/*
	 * suffixes of variables in this method
	 * *_p ... proposal
	 * *_o ... old value
	 * *_r ... reverse
	 */

	double cm = 0.5;
	double fraction_exchange = 0.03;
	double fraction_reincarnation = 0.05;
	double fraction_birth = 0.2;
	double fraction_death = 0.1;
	double fraction_tr_birth = 0.01;
	double fraction_tr_death = 0.01;
	long double alpha, MH_ratio;
	double epsilon_p;
	double u,ms;
	int n = M->n;
	double* theta_p = get_double_array(n);
	memcpy( theta_p, theta, sizeof(double)*n);
	model* M_p = malloc( sizeof(model) );
	int i,j,k = 0;
	long double posterior_o, posterior_p;
	double msp, msp_r, tp, tp_r;
	int accepted = 0;

	int si;
	int n_edges = 0;
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( M->P[i][j] ) {
				n_edges++;
			}
		}
	}

	posterior_o = compute_unnomr_posterior(eps, M, theta, D, N_u );
	printf("posterior_o %.20Lg\n", posterior_o);

	clone_poset( M, M_p );
	M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
	//parents_dt(&M);
	children_dt(M_p);
	epsilon_p = eps;

	FILE *samplelog;
	if ( (samplelog = fopen("data/output/SYNrun10.3sample", "w")) == NULL ) {
		exit(1);
	}

	for( k=0;k<burn_in+number_samples;k++) {

		ms = gsl_ran_flat (RNG, 0, 1);

		if( ms < cm ) {
			//relocate_whole_theta( theta, theta_p, n );

			si = rand()%n;;
			//si = k%n;
			relocate_theta_i( theta, theta_p, n, si );
			tp = compute_theta_transition_prob( theta, theta_p, n );
			tp_r = compute_theta_transition_prob( theta_p, theta, n );
			msp = cm;
			msp_r = cm;
		}
		else {
			if ( ms < cm+fraction_exchange ) {

				int ti,tj;
				propose_event_exchange_move( M_p, &tp, &ti, &tj );
				msp = msp_r = fraction_exchange;
				relocate_theta_i( theta, theta_p, n, ti );
				tp = compute_theta_transition_prob( theta, theta_p, n );
				tp_r = compute_theta_transition_prob( theta_p, theta, n );
				relocate_theta_i( theta, theta_p, n, tj );
				tp = tp*compute_theta_transition_prob( theta, theta_p, n );
				tp_r = tp_r*compute_theta_transition_prob( theta_p, theta, n );
				//tp = tp_r = 1;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation ) {
				if ( !propose_reincarnation_move( M_p, D, N_u ) ) {
					k--;
					continue;
				}
				msp = msp_r = fraction_reincarnation;
				tp = tp_r = 1;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth ) {
				//TODO: ugly: handle the case where no more cover relations are possible nicer!
				if( !propose_new_cover_relation(M_p, &tp, D, N_u ) ) {
					k--;
					continue;
				}
				tp_r = get_tp_for_remove_cover_move(M_p);
				msp = fraction_birth;
				msp_r = fraction_death;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death ) {
				if( !propose_delete_cover_relation(M_p, &tp ) ) {
					k--;
					continue;
				}
				tp_r = get_tp_for_new_cover_move(M_p, D, N_u );
				msp_r = fraction_birth;
				msp = fraction_death;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death+fraction_tr_birth ) {
				if( !propose_new_transitive_closure_relation(M_p, &tp, D, N_u ) ) {
					k--;
					continue;
				}
				tp_r = get_tp_for_delete_transitive_closure_relation_move(M_p);
				msp = fraction_tr_birth;
				msp_r = fraction_tr_death;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death+fraction_tr_birth+fraction_tr_death ) {
				if( !propose_delete_transitive_closure_relation(M_p, &tp, D, N_u ) ) {
					k--;
					continue;
				}
				tp_r = get_tp_for_new_transitive_closure_relation_move(M_p);
				msp = fraction_tr_death;
				msp_r = fraction_tr_birth;
			}
			else {
				relocate_epsilon( eps, &epsilon_p );
				tp = tp_r = get_tp_epsilon_relocation( epsilon_p );
				msp = msp_r = 1 - (cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death);
			}
			//just for testing perturb theta after structure move
			for ( i=0;i<2;i++ ) {
				si = rand()%n;
				//si = k%n;
				relocate_theta_i( theta, theta_p, n, si );
				tp = tp*compute_theta_transition_prob( theta, theta_p, n );
				tp_r = tp_r*compute_theta_transition_prob( theta_p, theta, n );
			}
			free_lattice_children( M_p );
			M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
			//parents_dt(&M);
			children_dt(M_p);
		}

		posterior_p = compute_unnomr_posterior(epsilon_p, M_p, theta_p, D, N_u );
		//printf("posterior_p %.20Lg\n", posterior_p);


		MH_ratio = posterior_p * msp_r * tp_r / (posterior_o * msp * tp );

		alpha = MIN(1,MH_ratio);

		u = gsl_ran_flat (RNG, 0, 1);

		if ( alpha > u ) {
			posterior_o = posterior_p;
			if( ms < cm ) {
				memcpy( theta, theta_p, sizeof(double)*n);
			}
			else {
				int h=0;
				for( i=0;i<n;i++ ) {
					for( j=0;j<n;j++ ) {
						if( M->P[i][j] != M_p->P[i][j] ) {
							if( M_p->P[i][j] ) {
								printf("%d+%d  ", i+1, j+1);
							}
							else {
								printf("%d-%d  ", i+1, j+1);
							}
						}
						h += M->P[i][j];
					}
				}
				if ( h != n_edges ) {
					printf("AAAALLLLEEEERRRRTTTTT");
				}
				free_poset( M );
				free_lattice_children( M );
				clone_poset( M_p, M );
				M->J_P = bfs_order_ideals(M->P, M->n, &(M->m), M->lin_ext);
				//parents_dt(M);
				children_dt(M);

				eps = epsilon_p;
				if ( ms < cm+fraction_exchange ) {
					printf("successful poset event EXCHANGE move!!!! posterior %.10Lg number edges %d epsilon: %.3f\n", posterior_o, n_edges, eps);
				}
				else if ( ms < cm+fraction_exchange+fraction_reincarnation ) {
					printf("successful poset REINCARNATION move!!!! posterior %.10Lg number edges %d epsilon: %.3f\n", posterior_o, n_edges, eps);
				}
				else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth ) {
					n_edges++;
					printf("successful poset BIRTH move!!!! posterior %.10Lg number edges %d epsilon: %.3f\n", posterior_o, n_edges, eps);
				}
				else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death ) {
					n_edges--;
					printf("successful poset DEATH move!!!! posterior %.10Lg number edges %d epsilon: %.3f\n", posterior_o, n_edges, eps);
				}
				else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death+fraction_tr_birth ) {
					n_edges = 0;
					for( i=0;i<n;i++ ) {
						for( j=0;j<n;j++ ) {
							if( M->P[i][j] ) {
								n_edges++;
							}
						}
					}
					printf("successful transitive closure BIRTH move!!!! posterior %.10Lg number edges %d epsilon: %.3f\n", posterior_o, n_edges, eps);
				}
				else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death+fraction_tr_birth+fraction_tr_death ) {
					n_edges = 0;
					for( i=0;i<n;i++ ) {
						for( j=0;j<n;j++ ) {
							if( M->P[i][j] ) {
								n_edges++;
							}
						}
					}
					printf("successful transitive closure DEATH move!!!! posterior %.10Lg number edges %d epsilon: %.3f\n", posterior_o, n_edges, eps);
				}
			}


			if( k>burn_in ) {
				accepted ++;
				//if ( !(accepted%record_ith) ) {
				//if ( !(k%1000) ) {
				//printf("sample number %d: posterior %.10Lg number edges %d\n", k,posterior_o, n_edges);
				//if( ms < cm ) {
				//print_model( M );
				//printf("theta: ");
				//print_double_array( theta, n );
				//}
				//else
				//{
				//printf("poset:\n");
				//print_int_matrix( M->P, n, n);
				//}
				//write_poset(0, "data/output/SYNrun1", M->P, n, k);
				//int** T = get_int_matrix(n, n);
				//transitive_closure( M->P, T, n );
				//write_poset(0, "data/output_tc/SYNrun1", T, n, k);
				//free_int_matrix( T, n );

				//}
			}
		}
		else {
			memcpy( theta_p, theta, sizeof(double)*n);

			free_poset( M_p );
			free_lattice_children( M_p );
			clone_poset( M, M_p );
			M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
			//parents_dt(M_p);
			children_dt(M_p);
			epsilon_p = eps;
		}

		//log samples
		if( k>burn_in && !(k%record_ith) ) {
			for (j=0; j<n; j++) {
				fprintf(samplelog, DOUBLE_FORMAT, theta[j]);
				fprintf(samplelog, "\t");
			}
			fprintf(samplelog, "posterior %.10Lg log-posterior %.5Lg number edges %d epsilon: %.3f accepted: %d, samplenr: %d u: %.5f ms: %.5f\n", posterior_o, log10l(posterior_o), n_edges, eps, accepted, k, u, ms );
			//write_poset(0, "data/output/SYNrun8.1", M->P, n, k);
			//int** T = get_int_matrix(n, n);
			//transitive_closure( M->P, T, n );
			//write_poset(0, "data/output_tc/SYNrun8.1", T, n, k);
			//free_int_matrix( T, n );
			if ( !(k%1000) ) {
				printf("sample number %d: posterior %.10Lg number edges %d number order ideals %d\n", k,posterior_o, n_edges, M_p->m);
			}
		}

	}

	fclose(samplelog);

	free(M_p);
	printf("poset:\n");
	print_int_matrix( M->P, n, n);
	printf("\n accepted: %d, ratio %f\n", accepted, (float)accepted/number_samples);
}

/**
 * Expectation over theta (i.e. P[M|D)
 * @param eps
 * @param M
 * @param theta
 * @param D
 * @param N_u
 * @param burn_in
 * @param number_samples
 * @param record_ith
 * @return
 */
long double start_Exp_theta_MH( double eps, model* M, double* theta, data* D, int N_u, int burn_in, int number_samples, int record_ith ) {

	int i,j,k = 0;
	long double posterior_o, posterior_p, posterior_sum;
	posterior_sum = 0;
	double msp, msp_r, tp, tp_r, u;
	int accepted = 0;
	int recorded = 0;
	int si;
	int n = M->n;
	double* theta_p = get_double_array(n);
	memcpy( theta_p, theta, sizeof(double)*n);
	long double alpha, MH_ratio;

	posterior_o = compute_unnomr_posterior(eps, M, theta, D, N_u );

	for( k=0;k<burn_in+number_samples;k++) {

		si = rand()%n;;
		//si = k%n;
		relocate_theta_i( theta, theta_p, n, si );
		tp = compute_theta_transition_prob( theta, theta_p, n );
		tp_r = compute_theta_transition_prob( theta_p, theta, n );
		msp = 1;
		msp_r = 1;
		posterior_p = compute_unnomr_posterior(eps, M, theta_p, D, N_u );
		//printf("posterior_p %.20Lg\n", posterior_p);


		MH_ratio = posterior_p * msp_r * tp_r / (posterior_o * msp * tp );

		alpha = MIN(1,MH_ratio);

		u = gsl_ran_flat (RNG, 0, 1);

		if ( alpha > u ) {
			memcpy( theta, theta_p, sizeof(double)*n);

			posterior_o = posterior_p;

		}
		else {
			memcpy( theta_p, theta, sizeof(double)*n);

		}
		if( k>burn_in ) {
			if ( !(k%record_ith) ) {
				recorded++;
				posterior_sum += posterior_o;
			}
		}

	}
	return posterior_sum/recorded;
}



void start_nested_MH(double eps, model* M, double* theta, data* D, int N_u, int burn_in, int number_samples, int record_ith ) {

	double fraction_exchange = 0.03;
	double fraction_reincarnation = 0.1;
	double fraction_birth = 0.5;
	double fraction_death = 0.22;
	long double alpha, MH_ratio;
	double epsilon_p;
	double u,ms;
	int n = M->n;
	double* theta_p = get_double_array(n);
	memcpy( theta_p, theta, sizeof(double)*n);
	model* M_p = malloc( sizeof(model) );
	int i,j,k = 0;
	long double posterior_o, posterior_p;
	double msp, msp_r, tp, tp_r;
	int accepted = 0;

	int si;
	int n_edges = 0;
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( M->P[i][j] ) {
				n_edges++;
			}
		}
	}

	FILE *samplelog;
	if ( (samplelog = fopen("data/output/SYNrun1sampleNest2", "w")) == NULL ) {
		exit(1);
	}
	epsilon_p = eps;
	posterior_o = start_Exp_theta_MH(eps, M, theta, D, N_u, 20000, 1000, 20 );;
	printf("posterior_o %.20Lg\n", posterior_o);

	clone_poset( M, M_p );
	M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
	//parents_dt(&M);
	children_dt(M_p);

	for( k=0;k<burn_in+number_samples;k++) {
		ms = gsl_ran_flat (RNG, 0, 1);

		if ( ms < fraction_exchange ) {

			int ti,tj;
			propose_event_exchange_move( M_p, &tp, &ti, &tj );
			msp = msp_r = fraction_exchange;
			relocate_theta_i( theta, theta_p, n, ti );
			tp = compute_theta_transition_prob( theta, theta_p, n );
			tp_r = compute_theta_transition_prob( theta_p, theta, n );
			relocate_theta_i( theta, theta_p, n, tj );
			tp = tp*compute_theta_transition_prob( theta, theta_p, n );
			tp_r = tp_r*compute_theta_transition_prob( theta_p, theta, n );
			//tp = tp_r = 1;
		}
		else if ( ms < fraction_exchange+fraction_reincarnation ) {
			if ( !propose_reincarnation_move( M_p, D, N_u ) ) {
				continue;
			}
			msp = msp_r = fraction_reincarnation;
			tp = tp_r = 1;
		}
		else if ( ms < fraction_exchange+fraction_reincarnation+fraction_birth ) {
			//TODO: ugly: handle the case where no more cover relations are possible nicer!
			if( !propose_new_cover_relation(M_p, &tp, D, N_u ) ) {
				continue;
			}
			tp_r = get_tp_for_remove_cover_move(M_p);
			msp = fraction_birth;
			msp_r = 1-(fraction_exchange+fraction_reincarnation+fraction_birth);
		}
		else if ( ms < fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death ){
			if( !propose_delete_cover_relation(M_p, &tp ) ) {
				continue;
			}
			tp_r = get_tp_for_new_cover_move(M_p, D, N_u );
			msp_r = fraction_birth;
			msp = fraction_death;
		}
		else {
			relocate_epsilon( eps, &epsilon_p );
			tp = tp_r = get_tp_epsilon_relocation( epsilon_p );
			msp = msp_r = 1 - (fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death);
		}

		//just for testing perturb theta after structure move
		for ( i=0;i<3;i++ ) {
			si = rand()%n;
			//si = k%n;
			relocate_theta_i( theta, theta_p, n, si );
			tp = tp*compute_theta_transition_prob( theta, theta_p, n );
			tp_r = tp_r*compute_theta_transition_prob( theta_p, theta, n );
		}
		free_lattice_children( M_p );
		M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
		//parents_dt(&M);
		children_dt(M_p);

		posterior_p = start_Exp_theta_MH(epsilon_p, M_p, theta_p, D, N_u, 1000, 1000, 20 );
		//printf("posterior_p %.20Lg\n", posterior_p);


		MH_ratio = posterior_p * msp_r * tp_r / (posterior_o * msp * tp );

		alpha = MIN(1,MH_ratio);

		u = gsl_ran_flat (RNG, 0, 1);

		if ( alpha > u ) {
			posterior_o = posterior_p;

			for( i=0;i<n;i++ ) {
				for( j=0;j<n;j++ ) {
					if( M->P[i][j] != M_p->P[i][j] ) {
						printf("%d %d  ", i+1, j+1);
					}
				}
			}
			free_poset( M );
			free_lattice_children( M );
			clone_poset( M_p, M );
			M->J_P = bfs_order_ideals(M->P, M->n, &(M->m), M->lin_ext);
			//parents_dt(M);
			children_dt(M);

			eps = epsilon_p;
			if ( ms < fraction_exchange ) {
				printf("successful poset event EXCHANGE move!!!! posterior %.10Lg number edges %d\n", posterior_o, n_edges);
			}
			else if ( ms < fraction_exchange+fraction_reincarnation ) {
				printf("successful poset REINCARNATION move!!!! posterior %.10Lg number edges %d\n", posterior_o, n_edges);
			}
			else if ( ms < fraction_exchange+fraction_reincarnation+fraction_birth ) {
				n_edges++;
				printf("successful poset BIRTH move!!!! posterior %.10Lg number edges %d\n", posterior_o, n_edges);
			}
			else if ( ms < fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death ) {
				n_edges--;
				printf("successful poset DEATH move!!!! posterior %.10Lg number edges %d\n", posterior_o, n_edges);
			}


			if( k>burn_in ) {
				accepted ++;
				if ( !(accepted%record_ith) ) {
					printf("sample number %d: posterior %.10Lg number edges %d\n", k,posterior_o, n_edges);

					//printf("poset:\n");
					//print_int_matrix( M->P, n, n);
					//write_poset(0, "data/output/run16", M->P, n, k);
					//int** T = get_int_matrix(n, n);
					//transitive_closure( M->P, T, n );
					//write_poset(0, "data/output_tc/run16", T, n, k);
					//free_int_matrix( T, n );

				}
			}
		}
		else {

			free_poset( M_p );
			free_lattice_children( M_p );
			clone_poset( M, M_p );
			M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
			//parents_dt(M_p);
			children_dt(M_p);
			epsilon_p = eps;

		}

		//log samples
		if( k>burn_in && !(k%record_ith) ) {
			fprintf(samplelog, "posterior %.10Lg log-posterior %.5Lg number edges %d epsilon: %.3f\n", posterior_o, log10l(posterior_o), n_edges, eps );
			write_poset(0, "data/output/SYNrun1Nest2", M->P, n, k);
			int** T = get_int_matrix(n, n);
			transitive_closure( M->P, T, n );
			write_poset(0, "data/output_tc/SYNrun1Nest2", T, n, k);
			free_int_matrix( T, n );
		}
	}
	free(M_p);
	printf("poset:\n");
	print_int_matrix( M->P, n, n);
	printf("\n accepted: %d, ratio %f\n", accepted, (float)accepted/number_samples);

}

void run_MH_sampler( model* M, double *theta_in, double epsilon_in, data* D, int N_u, int number_samples, int thinout,
		double **theta_matrix_out, int ***edges_cube_out, double *epsilon_out, double *log_posterior_out ) {


	/*
	 * suffixes of variables in this method
	 * *_p ... proposal
	 * *_o ... old value
	 * *_r ... reverse
	 */
	double cm = 0.5;
	double fraction_exchange = 0.03;
	double fraction_reincarnation = 0.05;
	double fraction_birth = 0.2;
	double fraction_death = 0.1;
	double fraction_tr_birth = 0.01;
	double fraction_tr_death = 0.01;
	long double alpha, MH_ratio;
	double epsilon_p;
	double u,ms;
	int n = M->n;
	double* theta_p = get_double_array(n);
	memcpy( theta_p, theta_in, sizeof(double)*n);
	model* M_p = malloc( sizeof(model) );
	int i,j,k = 0;
	long double posterior_o, posterior_p;
	double msp, msp_r, tp, tp_r;
	int accepted = 0;

	int si;
	int n_edges = 0;
	for( i=0;i<n;i++ ) {
		for( j=0;j<n;j++ ) {
			if( M->P[i][j] ) {
				n_edges++;
			}
		}
	}
	posterior_o = compute_unnomr_posterior(epsilon_in, M, theta_in, D, N_u );

	clone_poset( M, M_p );
	M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
	//parents_dt(&M);
	children_dt(M_p);
	epsilon_p = epsilon_in;

	for( k=0;k<number_samples*thinout;k++) {
		ms = gsl_ran_flat (RNG, 0, 1);

		if( ms < cm ) {
			//relocate_whole_theta( theta, theta_p, n );
			//printf("propose relocate theta\n");
			si = rand()%n;;
			//si = k%n;
			relocate_theta_i( theta_in, theta_p, n, si );
			tp = compute_theta_transition_prob( theta_in, theta_p, n );
			tp_r = compute_theta_transition_prob( theta_p, theta_in, n );
			msp = cm;
			msp_r = cm;
		}
		else {
			if ( ms < cm+fraction_exchange ) {
				//printf("propose event exchange\n");
				int ti,tj;
				propose_event_exchange_move( M_p, &tp, &ti, &tj );
				msp = msp_r = fraction_exchange;
				relocate_theta_i( theta_in, theta_p, n, ti );
				relocate_theta_i( theta_in, theta_p, n, tj );
				tp = compute_theta_transition_prob( theta_in, theta_p, n );
				tp_r = compute_theta_transition_prob( theta_p, theta_in, n );
				//tp = tp_r = 1;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation ) {
				//printf("propose reincarnation\n");
				if ( !propose_reincarnation_move( M_p, D, N_u ) ) {
					k--;
					continue;
				}
				msp = msp_r = fraction_reincarnation;
				tp = tp_r = 1;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth ) {
				//TODO: ugly: handle the case where no more cover relations are possible nicer!
				//printf("propose new cover\n");
				if( !propose_new_cover_relation(M_p, &tp, D, N_u ) ) {
					k--;
					continue;
				}
				tp_r = get_tp_for_remove_cover_move(M_p);
				msp = fraction_birth;
				msp_r = fraction_death;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death ) {
				//printf("propose delete cover\n");
				if( !propose_delete_cover_relation(M_p, &tp ) ) {
					k--;
					continue;
				}
				tp_r = get_tp_for_new_cover_move(M_p, D, N_u );
				msp_r = fraction_birth;
				msp = fraction_death;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death+fraction_tr_birth ) {
				//printf("propose new tc\n");
				if( !propose_new_transitive_closure_relation(M_p, &tp, D, N_u ) ) {
					k--;
					continue;
				}
				tp_r = get_tp_for_delete_transitive_closure_relation_move(M_p);
				msp = fraction_tr_birth;
				msp_r = fraction_tr_death;
			}
			else if ( ms < cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death+fraction_tr_birth+fraction_tr_death ) {
				//printf("propose delete tc\n");
				if( !propose_delete_transitive_closure_relation(M_p, &tp, D, N_u ) ) {
					k--;
					continue;
				}
				tp_r = get_tp_for_new_transitive_closure_relation_move(M_p);
				msp = fraction_tr_death;
				msp_r = fraction_tr_birth;
			}
			else {
				//printf("propose relocate epsilon\n");
				relocate_epsilon( epsilon_in, &epsilon_p );
				tp = tp_r = get_tp_epsilon_relocation( epsilon_p );
				msp = msp_r = 1 - (cm+fraction_exchange+fraction_reincarnation+fraction_birth+fraction_death);
			}
			if ( ms > cm+fraction_exchange ) {
				//just for testing perturb theta after structure move
				for ( i=0;i<2;i++ ) {
					si = rand()%n;
					//si = k%n;
					relocate_theta_i( theta_in, theta_p, n, si );
					tp = tp*compute_theta_transition_prob( theta_in, theta_p, n );
					tp_r = tp_r*compute_theta_transition_prob( theta_p, theta_in, n );
				}
			}
			free_lattice_children( M_p );
			M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
			//parents_dt(&M);
			children_dt(M_p);
		}

		posterior_p = compute_unnomr_posterior(epsilon_p, M_p, theta_p, D, N_u );


		MH_ratio = posterior_p * msp_r * tp_r / (posterior_o * msp * tp );

		alpha = MIN(1,MH_ratio);

		u = gsl_ran_flat (RNG, 0, 1);

		if ( alpha > u ) {
			posterior_o = posterior_p;
			if( ms < cm ) {
				memcpy( theta_in, theta_p, sizeof(double)*n);
			}
			else {

				free_poset( M );
				//printf("before free lattice children M\n");
				free_lattice_children( M );
				clone_poset( M_p, M );
				M->J_P = bfs_order_ideals(M->P, M->n, &(M->m), M->lin_ext);
				//parents_dt(M);
				children_dt(M);

				epsilon_in = epsilon_p;
			}
			accepted ++;
		}
		else {
			memcpy( theta_p, theta_in, sizeof(double)*n);

			free_poset( M_p );
			//printf("before free lattice children M_p\n");
			free_lattice_children( M_p );
			clone_poset( M, M_p );
			M_p->J_P = bfs_order_ideals(M_p->P, M_p->n, &(M_p->m), M_p->lin_ext);
			//parents_dt(M_p);
			children_dt(M_p);
			epsilon_p = epsilon_in;
		}

		//log samples
		if( !(k%thinout) ) {
			memcpy( theta_matrix_out[k/thinout], theta_in, sizeof(double)*n);
			for(i=0;i<n;i++ ) {
				memcpy( edges_cube_out[k/thinout][i], M->P[i], sizeof(int)*n );
			}
			epsilon_out[k/thinout] = epsilon_in;
			log_posterior_out[k/thinout] = (double)log10l(posterior_o);
			//printf("posterior %.10Lg log-posterior %.5Lg epsilon: %.3f accepted: %d, samplenr: %d u: %.5f ms: %.5f\n", posterior_o, log10l(posterior_o), epsilon_in, accepted, k, u, ms );
			/*n_edges = 0;
			for( i=0;i<n;i++ ) {
				for( j=0;j<n;j++ ) {
					if( M->P[i][j] ) {
						n_edges++;
					}
				}
			}
			fprintf(samplelog, "posterior %.10Lg log-posterior %.5Lg number edges %d epsilon: %.3f accepted: %d, samplenr: %d u: %.5f ms: %.5f\n", posterior_o, log10l(posterior_o), n_edges, epsilon_in, accepted, k, u, ms );
			 */
		}

	}
	free_poset( M_p );
	free_lattice_children( M_p );
	free(M_p);
}


#endif /* BAYES_DT_CBN_H_ */
