/*
 * bdt_cbn.c
 *
 *  Created on: Jan 3, 2012
 *      Author: Thomas Sakoparnig
 */

#include "bayes_dt_cbn.h"

int main(int argc, char **argv)
{
	double eps = 0.0;  // e
	char *filestem = '\0';
	unsigned int seed = (unsigned) time(NULL);  // r, random seed
	int f_flag = 0;

	int c = 0;
	while((c = getopt(argc, argv, "M:B:R:S:e:N:f:r:XvhP")) != EOF )
	{
		switch(c)
		{
		case 'f':
			filestem = optarg;
			f_flag++;
			break;
		}

	}
	if (! f_flag)
	{
		printf("Error: No input file specified!\n");
		exit(1);
	}

	srand(seed);
	RNG = gsl_rng_alloc (gsl_rng_taus);  // global variable
	gsl_rng_set(RNG, seed);  // seed rng

	int i, k;

	model M;
	read_poset_dt(filestem, &M);

	// precompute binary expansions
	precompute_binary(M.n);

	M.lin_ext = get_int_array(M.n);  // a linear extension of the poset
	double* theta = get_double_array(M.n);  // mutation probabilities


	int N, N_u, burn_in;
	int** pat = read_patterns(filestem, &N, M.n);
	int* pat_idx = get_int_array(N);
	data* D = make_data_set(pat, N, M.n, &N_u, pat_idx);
	for (k=0; k<N; k++)
		free(pat[k]);
	free(pat);

	for( i=0;i<M.n;i++)
		theta[i] = 0.4; //initial values for MH

	//	theta[0] = 0.76180939;
	//	theta[1] = 0.830335602;
	//	theta[2] = 0.5008396168;
	//	theta[3] = 0.8149854961;
	//	theta[4] = 0.3818451012;
	//	theta[5] = 0.275943532;
	//	theta[6] = 0.7349472765;
	//	theta[7] = 0.8992577086;
	//	theta[8] = 0.675706852;
	//	theta[9] = 0.1117242475;

	eps = 0.01; //initial value for MH

	int number_samples = 500000;
	int record_ith = 1;
	burn_in = 0;


	//just for testing:!!!
	// Inititalize data structures
	M.J_P = bfs_order_ideals(M.P, M.n, &(M.m), M.lin_ext);
	//parents_dt(&M);
	children_dt(&M);
	//construct_sublattices(D, N_u, &M);
	compatibility(D, N_u, &M);

	double* Prob = get_double_array(M.m);
	double* theta_exit = get_double_array(M.m);
	compute_theta_exit( theta, &M, theta_exit);
	compute_all_cbn_prob( theta, &M, Prob, theta_exit );
	double** cond_err_prob = get_double_matrix(M.m, N_u);
	compute_cond_error_prob( &M, D, N_u, eps, cond_err_prob );

	//print_data( D, N_u, M.n, M.m);
	/*	long double lik;
	for( i=0;i<10;i++ ) {
		lik = compute_likelihood(eps, &M, theta, D, N_u, Prob, cond_err_prob );
	}
	printf("%.20Lg\n",lik);
	long double posterior;
	posterior = compute_unnomr_posterior( eps, &M, theta, D, N_u );
	printf("posterior %.20Lg\n", posterior);
	 */

	/*	long double posterior = start_Exp_theta_MH(eps, &M, theta, D, N_u, 10000, 5000, 50 );
	printf("posterior %.20Lg\n", posterior);

	theta[2]= 0.8;
	theta[7] = 0.2;
	theta[9] = 0.4;
	posterior = start_Exp_theta_MH(eps, &M, theta, D, N_u, 1000, 1000, 20 );
	printf("2. run posterior %.20Lg\n", posterior);
	 */
	start_MH( eps, &M, theta, D, N_u, burn_in, number_samples, record_ith );
	//start_nested_MH( eps, &M, theta, D, N_u, burn_in, number_samples, record_ith );
	return 0;
}

/**
 *
 * @param theta_in array[n] of starting thetas
 * @param nevents number of events
 * @param epsilon_in starting epsilon
 * @param edges_in array[n*n] of starting edges (poset)
 * @param number_smaples number of desired returned samples
 * @param thinout number of samples to discard between recording one
 * @param data array[n*ncases] data for inference
 * @param ncases number of cases i.e. data rows
 * @param theta_out array[n*number_samples] thetas sampled
 * @param epsilon_out array[number_samples] epsilons sampled
 * @param edges_out array[n*n*k] edges sampled (poset)
 */
void sample_full_cbn(double *theta_in, int *nevents, double *epsilon_in, int * edges_in, int *number_samples, int *thinout, int *patdata,
		int *number_cases, double *theta_out, double *epsilon_out, int* edges_out, double* log_posterior_out ) {

	unsigned int seed = (unsigned) time(NULL);  // r, random seed
	srand(seed);
	RNG = gsl_rng_alloc (gsl_rng_taus);  // global variable
	gsl_rng_set(RNG, seed);  // seed rng

	int i,j,k,n;
	n=(*nevents);
	int n_samples = (*number_samples);
	int n_cases = (*number_cases);


	model M;
	M.n = n;
	M.P = get_int_matrix(n, n);

	for( i=0;i<n;i++){
		for( j=0;j<n;j++) {
			M.P[i][j] = edges_in[i*n+j];
		}
	}
	precompute_binary(M.n);

	M.lin_ext = get_int_array(M.n);  // a linear extension of the poset
	int** pat = get_int_matrix( n_cases,n );
	for( i=0;i<n_cases;i++) {
		for( j=0;j<n;j++ ) {
			pat[i][j] = patdata[i*n+j];
		}
	}

	int* pat_idx = get_int_array(n_cases);
	int N_u;
	data* D = make_data_set(pat, n_cases, M.n, &N_u, pat_idx);

	for (k=0; k<n_cases; k++)
		free(pat[k]);
	free(pat);

	double ** theta_matrix_out = get_double_matrix( n_samples ,n );
	int *** edges_cube_out = get_int_cube( n_samples, n, n );
	for( i=0;i<n_samples;i++ ) {
		for( j=0;j<n;j++ ) {
			for( k=0;k<n;k++ ) {
				edges_cube_out[i][j][k] = 0;
			}
		}
	}

	M.J_P = bfs_order_ideals(M.P, M.n, &(M.m), M.lin_ext);
	//parents_dt(&M);
	children_dt(&M);
	run_MH_sampler( &M, theta_in, (*epsilon_in), D, N_u, n_samples, (*thinout), theta_matrix_out, edges_cube_out, epsilon_out, log_posterior_out );

	for( i=0;i<n_samples;i++) {
		for( j=0;j<n;j++ ) {
			theta_out[i*n+j] = theta_matrix_out[i][j];
		}
	}
	for( i=0;i<n_samples;i++ ) {
		for( j=0;j<n;j++ ) {
			for( k=0;k<n;k++ ) {
				//printf("%d ",i*n*n+j*n+k);
				edges_out[i*n*n+j*n+k] = edges_cube_out[i][j][k];
				//printf(".");
				//printf("%d ", edges_cube_out[i][j][k]);
			}
		}
	}
	free_double_matrix( theta_matrix_out, n_samples );
	free_int_cube( edges_cube_out, n_samples, n );
	//free_data( D, N_u, n);
}
